// --------------------------------------------------------------------
// EDA385 - FPGA POD'
//   => SD/SDHC driver
//---------------------------------------------------------------------

//#include "stdio.h"
#include "SD_Driver.h"
#include "xparameters.h"
#include "xuartlite.h"
#include "string.h"
#include "xio.h"

	 // Global Variables
  
	 Xboolean gbActive = XFALSE;
	 Xuint8 gszCSD[17];
	 Xuint8 gszCID[17];
    Xuint8 read_status;
    Xuint8 response_buffer[20];
    Xuint8 RCA[2];
    Xuint8 cmd_buffer[5];
	 Xuint32 SD_gpio_o = 0;
	 Xuint32 SD_gpio_3states = 0xffffffff;
	 
    
  // Command Constants (with start and stop bits)
    
    const Xuint8 cmd0[5]   = {0x40,0x00,0x00,0x00,0x00};   // SD Initialization
    const Xuint8 cmd55[5]  = {0x77,0x00,0x00,0x00,0x00};   // App. Specific cmd
    const Xuint8 cmd2[5]   = {0x42,0x00,0x00,0x00,0x00};   // All Send CID
    const Xuint8 cmd3[5]   = {0x43,0x00,0x00,0x00,0x00};   // Publish new RCA (relative adress)
    const Xuint8 cmd7[5]   = {0x47,0x00,0x00,0x00,0x00};   // Sel/Desel SD (multiple cards systems)
    const Xuint8 cmd9[5]   = {0x49,0x00,0x00,0x00,0x00};   // Ask Card Specific Data (CSD)
    const Xuint8 cmd10[5]  = {0x4a,0x00,0x00,0x00,0x00};   // Ask CID (Card ID)
    const Xuint8 cmd16[5]  = {0x50,0x00,0x00,0x02,0x00};   // Set Block Length to 512 bytes (mandatory with SDHC)
    const Xuint8 cmd17[5]  = {0x51,0x00,0x00,0x00,0x00};   // Read Single Block
    const Xuint8 acmd6[5]  = {0x46,0x00,0x00,0x00,0x02};   // Set Bus Width (1 or 4 bits DAT lines)
    //const Xuint8 acmd41[5] = {0x69,0x0f,0xf0,0x00,0x00};   // Ask card operating conditions
	 const Xuint8 acmd41_sdhc[5] = {0x69,0x40,0xff,0x00,0x00};   // Ask card operating conditions, wide voltage range, sdhc
	 const Xuint8 acmd41_sd[5] = {0x69,0x00,0xff,0x00,0x00};   // Ask card operating conditions, wide voltage range, sdhc
    const Xuint8 acmd51[5] = {0x73,0x00,0x00,0x00,0x00};   // Ask SD Configuration register
	 const Xuint8 cmd8[5]   = {0x48,0x00,0x00,0x01,0xAA};   // Check SDHC compatibility => 2.7-3.6V

//-------------------------------------------------------------------------

    void Ncr(void)
    {
      SD_CMD_IN();
      // 2 clocks period....
      SD_CLK_RESET();  
      SD_CLK_SET();
      SD_CLK_RESET();
      SD_CLK_SET();
    } 

    void Ncc(void)
    {
      int i;
      // The command response time is 0 to 8 bytes for SDC
      for(i=0;i<8;i++)
      {
        SD_CLK_RESET();
        SD_CLK_SET();
      }
    }
    
//-------------------------------------------------------------------------

    void dump_CID(Xuint8 szCID[])
    {
       int i = 1;
       xil_printf("SD-CARD CID:\r\n");
       xil_printf("  MID : %02Xh\r\n", szCID[i+0]);
       
     /* xil_printf("  OEM/Application ID(OLD):%02X%02Xh\r\n", szCID[i+2], szCID[i+1]);
       xil_printf("  Product Name(PNM):%C%C%C%C%C\r\n", szCID[i+3], szCID[i+4], szCID[i+5], szCID[i+6], szCID[i+7]);
       xil_printf("  Product Revision:%02Xh\r\n", szCID[i+8]);
       xil_printf("  Serial Number(PSN):%02X%02X%02X%02Xh\r\n", szCID[i+9], szCID[i+10], szCID[i+11], szCID[i+12]);
       xil_printf("  Manufacturer Date Code(MDT):%01X%02Xh\r\n", szCID[i+13] & 0x0F, szCID[i+14]);
       xil_printf("  CRC-7 Checksum(CRC7):%02Xh\r\n", szCID[i+15] >> 1);*/
        
    }

    Xboolean SD_GetCSD(Xuint8 szCSD[], Xuint8 len)
    {
        if (!gbActive)
            return XFALSE;
        if (len > 16)
            len = 16;        
        memcpy(szCSD, &gszCSD[1], len);         
        return XTRUE;        
    }

    Xboolean SD_GetCID(Xuint8 szCID[], Xuint8 len)
    {
        if (!gbActive)
            return XFALSE;
        if (len > 16)
            len = 16;        
        memcpy(szCID, &gszCID[1], len);         
        return XTRUE;        
    }

//-------------------------------------------------------------------------

  Xboolean SD_Init(void)
  {
      Xuint8 x,y;
      Xuint32 i;
		int try = 0;
		Xboolean SDHC = XFALSE;
		
	  // GPIO Initialization
       SD_CLK_OUT();

      // pull-high DAT3 to enter SD mode during at least 74 clocks or 1 ms...
      
       SD_DAT3_OUT();
       SD_DAT3_SET();
		 SD_CMD_OUT();
		 SD_CMD_SET();	
       
       for(i=0; i < 500000; i++)  // delay > 1 ms
          asm("nop");

		 SD_DAT0_IN();
       SD_CLK_SET();
       SD_DAT0_RESET();
	  
      // 1) Send CMD 0 
           
         gbActive = XFALSE;
         read_status=0;
         for(x=0;x<40;x++) Ncr();
         for(x=0;x<5;x++)  cmd_buffer[x]=cmd0[x];
         y = send_cmd(cmd_buffer);		
			
         // !!! THE CARD IS NOW IN IDENTIFICATION MODE !!!4
			
		// 1-bis) Send CMD8 for SDHC compatibility

          for(x=0;x<40;x++);
			 Ncc();   
          for(x=0;x<5;x++) 
				cmd_buffer[x]=cmd8[x];
          y = send_cmd(cmd_buffer);
          Ncr();
			 //xil_printf("\x1B[j");  // clear screen
          //xil_printf("cmd8...");
			 if(response_R(1)>1) 
				SDHC = XFALSE;
			 else
				SDHC = XTRUE;
			 
      // 2) Send CMD55 => Application command, wait for sd answer...  
       
        do
        {
          for(x=0;x<40;x++);
          //  After the last command has been sent, the host can continue 
          //  sending the next command after at least NCC clock periods. 
          Ncc();
			 
          for(x=0;x<5;x++) cmd_buffer[x]=cmd55[x];
          y = send_cmd(cmd_buffer);
          //xil_printf("\x1B[j");  // clear screen
          //xil_printf("cmd55...");
          //  After receiving the last card response, the host can start the 
          //  next command transmission after at least NRC clock cycles. This 
          //  timing is relevant for any host command. 
          Ncr();
          if(response_R(1)>1) //response too long or crc error
          { 
            xil_printf("error cmd55\r\n");
            return XFALSE;
          }
                   
          Ncc();
            
           // 3) Send CMD41 (ACMD41) => Operating condition of the sd card... 
        
            if (SDHC == XTRUE) 
				  for(x=0;x<5;x++) cmd_buffer[x]=acmd41_sdhc[x];
				else
				  for(x=0;x<5;x++) cmd_buffer[x]=acmd41_sd[x];	
				  
            y = send_cmd(cmd_buffer);
            //xil_printf("acmd41...");
            Ncr();			
       } while(response_R(3)==1);  // wait until the card is not busy
		
      // 4) send CMD2 and wait answer (ALL_SEND_CID)
      
        //xil_printf("\x1B[j"); 
		  Ncc();
        for(x=0;x<5;x++) cmd_buffer[x]=cmd2[x];
        y = send_cmd(cmd_buffer);
        //xil_printf("send cmd2\n");
        Ncr();
        if(response_R(2)>1) return XFALSE;
          
       // 5) CMD3 : asks the card to publish a new relative card address (RCA) 
       
        Ncc();
        for(x=0;x<5;x++) cmd_buffer[x]=cmd3[x];
        y = send_cmd(cmd_buffer);
        Ncr();
		  //xil_printf("send cmd3\n");
        if(response_R(6)>1) return XFALSE;
        RCA[0]=response_buffer[1];
        RCA[1]=response_buffer[2];
        
        
      // !!! THE CARD IS NOW IN DATA-TRANSFER MODE !!!
      
       // 6) CMD9 with given RCA : asks for card CSD register
       //    (block length, card storage capacity....)
       
        Ncc();
        for(x=0;x<5;x++) cmd_buffer[x]=cmd9[x];
        cmd_buffer[1] = RCA[0];
        cmd_buffer[2] = RCA[1];
        //xil_printf("send cmd9\n");		  
        y = send_cmd(cmd_buffer);
        Ncr();
        if(response_R(2)>1) return XFALSE;
        memcpy(gszCSD,&response_buffer[0] , sizeof(gszCSD));
        
       // 7) CMD10 : asks for card CID reg...
       
        Ncc();
        for(x=0;x<5;x++) cmd_buffer[x]=cmd10[x];
        cmd_buffer[1] = RCA[0];
        cmd_buffer[2] = RCA[1];  
		  //xil_printf("send cmd9\n");
        y = send_cmd(cmd_buffer);
        Ncr();
        if(response_R(2)>1) return XFALSE;
        memcpy(gszCID,&response_buffer[0] , sizeof(gszCID));    // richard add

       // 8) CMD7 : puts the card in the TRANSFER STATE  

        Ncc();
        for(x=0;x<5;x++) cmd_buffer[x]=cmd7[x];
        cmd_buffer[1] = RCA[0];
        cmd_buffer[2] = RCA[1];
		  //xil_printf("send cmd7\n");
        y = send_cmd(cmd_buffer);
        Ncr();
        if(response_R(1)>1) return XFALSE;
           
        // CMD16 : set block length to 512 bytes (in cmd16 array)
        
        Ncc();
        for(x=0;x<5;x++) cmd_buffer[x]=cmd16[x];
        y = send_cmd(cmd_buffer);  
        Ncr();
        if(response_R(1)>1) return XFALSE;
         
        //xil_printf("SD_card_init success\n");
        read_status = 1; // SD ready !!!
        gbActive = XTRUE;
        return XTRUE;
  }


  Xboolean SD_read_block(Xuint32 block_number, Xuint8 *buff)
  {
    //---------------------------------------------
    // !!!! buffer size must be 512 bytes !!!!  
    //---------------------------------------------
    
    Xuint8 c = 0; // data received
    Xuint32  i,addr;
    int try = 0;
    const int max_try = 5000;
    
      // 1) CMD17 : "Single Block Read",  parameter: block address 
      
        Ncc();
        addr = block_number * 512;
        cmd_buffer[0] = cmd17[0];
        cmd_buffer[1] = (addr >> 24 ) & 0xFF; // MSB
        cmd_buffer[2] = (addr >> 16 ) & 0xFF;
        cmd_buffer[3] = (addr >> 8 ) & 0xFF;
        cmd_buffer[4] = addr & 0xFF; // LSB
        send_cmd(cmd_buffer); 
        Ncr();
      
      
      // 2) get response
      
        while(1)
        {
          SD_CLK_RESET();
          SD_CLK_SET();
          if(!SD_DAT0_Read())  break;  // we start to receive data
          if (try++ > max_try) return XFALSE;        
        }
      
      // 3) read data (512byte = 1 block)
      
        for(i=0;i<512;i++)
        {
          Xuint8 j;
          for(j=0;j<8;j++)
          {
            SD_CLK_RESET();
            SD_CLK_SET();
            c <<= 1; 
            if(SD_DAT0_Read()) c |= 0x01;
          } 
          *buff=c;
          buff++;
        } 
      
        // CRC16 but we won't verify it (very baaaad !!!)
        
        for(i=0; i<16; i++)
        {
            SD_CLK_RESET();
            SD_CLK_SET();
        }
        
        read_status = 1;  //SD data next in
        return XTRUE;
  }

//-------------------------------------------------------------------------
//  Xuint8 response_R(Xuint8 s)
//    => read response from CMD line
//    => s : type of response (R1, R2, R3 or R6)
//-------------------------------------------------------------------------

  Xuint8 response_R(Xuint8 s)
  {
    Xuint8 a=0,b=0,c=0,r=0,crc=0;
    Xuint8 i,j=6,k;
	 Xuint32 res = 0;
    
    while(1)
    {
      SD_CLK_RESET();
      SD_CLK_SET();
		res = SD_CMD_Read();
      if(res == 0) break;
		if(crc++ >100)
		{
		  xil_printf("timeout...\r\n");
		  return 2;
		}
    } 
    
    crc = 0;
    if(s == 2) j = 17;  // for 136 bits R2 answer
  
    for(k=0; k < j; k++)
    {
      c = 0;
      if(k > 0)  b = response_buffer[k-1]; 
              
      for(i=0; i<8; i++)
      {
        SD_CLK_RESET();
        if(a > 0)
        c <<= 1; 
        else
        i++; 
        a++; 
        SD_CLK_SET();
        if(SD_CMD_Read()) 
		  {
		    c |= 0x01; // crc7 reception	 
		  }
        
        if(k > 0)  
        {
          crc <<= 1;
          if((crc ^ b) & 0x80)
          crc ^= 0x09;
          b <<= 1;
          crc &= 0x7f;
        }
      }
      
      if(s==3)   // CRC7 = 1111111
      { 
        if( k==1 &&(!(c&0x80)))  r=1;
      }
      
      response_buffer[k] = c;
    }
    
    if(s==1 || s==6)
    {
		if(c != ((crc<<1)+1))
			r=2; // computed and received crc7s don't match
    }
	 // when s=3, r=0, CRC is not checked
     
    return r; 
  }
  
//-------------------------------------------------------------------------
//  Xuint8 send_cmd(Xuint8 *in)
//    => send command and arguments on the CMD line, compute  crc7
//    => in : data to send 
//    => return computed CRC7
//-------------------------------------------------------------------------

  Xuint8 send_cmd(Xuint8 *in)
  {
    int i,j,z;
    Xuint8 b,crc=0;
    
    SD_CMD_OUT();			  
    
    // send 6 bytes (48 bits = CMD + arguments)
    
    for(i=0; i < 5; i++)
    {
      b = in[i];
      for(j=0; j<8; j++)
      {
        SD_CLK_RESET();
		  
		  for(z=0; z < 5000; z++) 
          asm("nop");

          if(b&0x80)
             SD_CMD_SET();	 
          else
             SD_CMD_RESET();

        crc <<= 1;
        SD_CLK_SET();
        if((crc ^ b) & 0x80)   // crc7 computation...
        crc ^= 0x09;
        
        b<<=1;
      } 
      crc &= 0x7f; 
    }  
    
    crc =((crc<<1)|0x01);
    b = crc;
    
    //  send CRC7
    for(j=0; j<8; j++)
    {
      SD_CLK_RESET();
      
        if(crc&0x80)
          SD_CMD_SET();
        else
          SD_CMD_RESET(); 
        
      SD_CLK_SET();
      
      crc<<=1;
    }  
     
    return b;  

	 
  }


//------------------------------------------------------------------------------
// Low-level functions
//------------------------------------------------------------------------------

    // Data directions
    
        void SD_DAT3_IN(void)
        {
          // DAT3 = bit3
          SD_gpio_3states |= 0x01; // set bit 3  (1 = input)
          XIo_Out32(XPAR_XPS_GPIO_0_BASEADDR + 4, SD_gpio_3states);
        }
        
        void SD_DAT3_OUT(void)
        {
          // DAT3 = bit3
          SD_gpio_3states &= ~0x01; // reset bit 3  (0 = outpout)
          XIo_Out32(XPAR_XPS_GPIO_0_BASEADDR + 4, SD_gpio_3states);
        }
        
        void SD_DAT0_IN(void)
        {
          // DAT0 = bit2
          SD_gpio_3states |= 0x02; // set bit 0  (1 = input)
          XIo_Out32(XPAR_XPS_GPIO_0_BASEADDR + 4, SD_gpio_3states);
        }
        
        void SD_DAT0_OUT(void)
        {
          // DAT0 = bit2
          SD_gpio_3states &= ~0x02; // reset bit 0 (0 = outpout)
          XIo_Out32(XPAR_XPS_GPIO_0_BASEADDR + 4, SD_gpio_3states);
        }
        
        void SD_CMD_IN(void)
        {
          // CMD = bit1
          SD_gpio_3states |= 0x04; // set bit 1  (1 = input)
          XIo_Out32(XPAR_XPS_GPIO_0_BASEADDR + 4, SD_gpio_3states);
        }
        
        void SD_CMD_OUT(void)
        {
          // CMD = bit1
          SD_gpio_3states &= ~0x04; // reset bit 1 (0 = outpout)
          XIo_Out32(XPAR_XPS_GPIO_0_BASEADDR + 4, SD_gpio_3states);
        }
        
        void SD_CLK_OUT(void)
        {
          // CLK = bit0
          SD_gpio_3states &= ~0x08; // reset bit 0 (0 = outpout)
          XIo_Out32(XPAR_XPS_GPIO_0_BASEADDR + 4, SD_gpio_3states);
        }

    // Set/Reset pins
    
        void SD_DAT3_SET(void)
        {
           SD_gpio_o |= 0x01;
           XIo_Out32(XPAR_XPS_GPIO_0_BASEADDR, SD_gpio_o);
        }
        
        void SD_DAT3_RESET(void)
        {
           SD_gpio_o &= ~0x01;
           XIo_Out32(XPAR_XPS_GPIO_0_BASEADDR, SD_gpio_o);;        
        }
        
        void SD_DAT0_SET(void)
        {
           SD_gpio_o |= 0x02;
           XIo_Out32(XPAR_XPS_GPIO_0_BASEADDR, SD_gpio_o);       
        }
        
        void SD_DAT0_RESET(void)
        {
           SD_gpio_o &= ~0x02;
           XIo_Out32(XPAR_XPS_GPIO_0_BASEADDR, SD_gpio_o);         
        }
        
        void SD_CMD_SET(void)
		  {
           SD_gpio_o |= 0x04;
           XIo_Out32(XPAR_XPS_GPIO_0_BASEADDR, SD_gpio_o);      
        }
        
        void SD_CMD_RESET(void)
        {
           SD_gpio_o &= ~0x04;
           XIo_Out32(XPAR_XPS_GPIO_0_BASEADDR, SD_gpio_o);          
        }
        
        void SD_CLK_SET(void)
        {
           SD_gpio_o |= 0x08;
           XIo_Out32(XPAR_XPS_GPIO_0_BASEADDR, SD_gpio_o);
        }
        
        void SD_CLK_RESET(void)
        {
           SD_gpio_o &= ~0x08;
           XIo_Out32(XPAR_XPS_GPIO_0_BASEADDR, SD_gpio_o);
        }
        
     // Read Functions
     
      unsigned int SD_DAT3_Read(void)
      {
           unsigned long Data_Read;
           Data_Read =  XIo_In32(XPAR_XPS_GPIO_0_BASEADDR); 
           if (Data_Read & 0x01)
           {
             return(1);
           }
           else 
             return(0);             
      }  
      
      unsigned int SD_DAT0_Read(void)
      {
           unsigned long Data_Read;
           Data_Read =  XIo_In32(XPAR_XPS_GPIO_0_BASEADDR);
           if (Data_Read & 0x02)
           {
             return(1);
           }
           else 
             return(0);                 
      }
      
      unsigned int SD_CMD_Read(void)
      {
           unsigned long Data_Read;
           Data_Read =  XIo_In32(XPAR_XPS_GPIO_0_BASEADDR);
           if (Data_Read & 0x04)
           {
				 return(1);
           }
           else 
			  {
				 return(0);
			  }
      }
