/*
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "stdio.h"
#include "stdlib.h"
#include "xbasic_types.h"
#include "xio.h"
#include "fat12.h"
#include "io.h"
#include "list.h"


// src have to be 512 byte or bigger
// prints the data in hexadecimal form
/*static void hexdump(void* src)
{
	Xuint8			*buff;
	Xuint32			rows;
	Xuint32			columns;
	
	buff = (Xuint8*)src;
	
	xil_printf("\r\n--- HEXDUMP ---\r\n");
	for (rows = 0; rows < 32; rows++) {
		xil_printf("%08x  ", rows*16);
		for (columns = 0; columns < 16; columns++) {
			xil_printf("%02x ", XIo_In8(rows*16+columns));
		}
		xil_printf("\r\n");
	}
}*/

static unsigned int load_int(char* ptr)
{
	unsigned int 	result;
	unsigned int 	tmp;
	
	result = 0x00000000;
	
	tmp = (unsigned int)XIo_In8(ptr);
	result |= tmp;
	tmp = (unsigned int)XIo_In8(ptr+1);
	tmp <<= 8;
	result |= tmp;
	tmp = (unsigned int)XIo_In8(ptr+2);
	tmp <<= 16;
	result |= tmp;
	tmp = (unsigned int)XIo_In8(ptr+3);
	tmp <<= 24;
	result |= tmp;
	
	return result;
}


// fails because MSB and stuff like that
// 00 02 is in a short 0x0200
static unsigned int load_short(char* ptr)
{	
	unsigned short 	result;
	unsigned short 	tmp;
	
	result = 0x0000;
	
	tmp = (unsigned short)XIo_In8(ptr);
	result |= tmp;
	tmp = (unsigned short)XIo_In8(ptr+1);
	tmp <<= 8;
	result |= tmp;
	
	return result;
}



void init_vbr()
{
	char					buffer[512];
	unsigned char		partition1_type;
	unsigned int		partition1_begin_sector;
	unsigned int		global_offset;
	char*					ptr;

	// read first block from io
	seek_from_start(0);
	read_from_io(buffer, 1, 512);

	partition1_begin_sector = load_int((buffer+446+8));
	xil_printf("\r\nPartition1 begin: 0x%08x\r\n", partition1_begin_sector);

	global_offset = partition1_begin_sector*512;

	seek_from_start(global_offset);
	read_from_io(buffer, 1, 512);

	/* VBR: volume boot record */
	fat32.bytes_per_sector = load_short(buffer+0x0b);
	xil_printf("bytes per sector: 0x%04x\r\n", fat32.bytes_per_sector);
	fat32.sectors_per_cluster = buffer[0x0d];
	fat32.reserved_sector_count = load_short(buffer+0x0e);
	fat32.number_of_fats = buffer[0x10];
	fat32.max_root_dir_entries = load_short(buffer+0x11);
	fat32.sectors_per_fat = load_short(buffer+0x16);
	fat32.fat_size = fat32.sectors_per_fat*fat32.number_of_fats;
	my_memcpy(fat32.fat_type, (buffer+0x36), 8);
	fat32.signature = load_short(buffer+0x1fe);

	/* At what sector does the fat table begin? */
	fat32.fat_base = partition1_begin_sector + fat32.reserved_sector_count;

	/* At what sector does the root directory start at? */
	fat32.root_dir_sector = fat32.fat_base + fat32.fat_size;

	/* At what address does the data region start at? */
	fat32.data_start_addr = fat32.root_dir_sector*fat32.bytes_per_sector + 
				fat32.max_root_dir_entries*32;
}

void print_vbr()
{
	int		i;

	xil_printf("Bytes per sector: \t\t %d\r\n", fat32.bytes_per_sector);
	xil_printf("Sector per cluster: \t\t %d\r\n", fat32.sectors_per_cluster);
	xil_printf("Reserved sector count: \t\t %d\r\n", fat32.reserved_sector_count);
	xil_printf("Number of FATs: \t\t %d\r\n", fat32.number_of_fats);
	xil_printf("Max root dir entries: \t\t %d\r\n", fat32.max_root_dir_entries);
	xil_printf("Sectors per fat: \t\t %d\r\n", fat32.sectors_per_fat);
	xil_printf("Fat size: \t\t\t %d\r\n", fat32.fat_size);
	xil_printf("Root dir sector: \t\t 0x%08x\r\n", fat32.root_dir_sector);
	xil_printf("Fat type: \t\t\t ");
	for (i = 0; i < 8; i++) {
		xil_printf("%c", fat32.fat_type[i]);
	}
	xil_printf("\r\n");
	xil_printf("Signature: \t\t\t 0x%02hx \r\n\r\n", fat32.signature);
	xil_printf("Fat base: \t\t\t %d\r\n", fat32.fat_base);
	xil_printf("Fat base addr: \t\t\t 0x%08x\r\n", fat32.fat_base*fat32.bytes_per_sector);
	xil_printf("Data start base address: \t 0x%08x\r\n", fat32.data_start_addr);
}

// look up the next cluster number in the FAT
// be aware that all the ugly code, masking and shifting, even/uneven is because that in FAT12
// each entry is 12 bit, 1.5 bytes! Straaange.
unsigned int get_next_cluster(unsigned int cluster_no)
{
	//unsigned char		fat_buf[100];
	unsigned short		fat_entry;
	unsigned int		entry_address;
	unsigned short		even_mask;
	unsigned short		uneven_mask;
	unsigned int		next_cluster_no;
	unsigned short		low_byte;
	unsigned short		high_byte;

	uneven_mask = 0xfff0;
	even_mask = 0x0fff;
	

	// the first cluster in the data region is cluster #2
	// therefore, the first 2 entries in the fat tables are unused
	// The FAT12 file system uses 12 bits per FAT entry, two entries spans 3 bytes
	entry_address = fat32.fat_base*fat32.bytes_per_sector + (cluster_no)*1.5;

	// read 1 entry + some junk into fat_entry. 4 bits aren't used
	seek_from_start(entry_address);
	read_from_io(&fat_entry, 1, 2);

	if (cluster_no % 2 == 0) {
		// get the entry we want
		low_byte = (fat_entry & 0xff00)>>8;
		high_byte = fat_entry & 0x00ff;
				
		high_byte &= 0x000f;
		high_byte <<= 8;

		next_cluster_no = high_byte | low_byte;
	} else {
		// get the entry we want
		low_byte = (fat_entry & 0x00ff)<<4;
		high_byte = (fat_entry & 0xff00)>>12;
		
		next_cluster_no = high_byte | low_byte;
	}

	return next_cluster_no;
}

// what memory address does the cluster start at?
unsigned int get_cluster_addr(unsigned int cluster_no)
{
	return (fat32.data_start_addr)+(cluster_no-2)*fat32.sectors_per_cluster*fat32.bytes_per_sector;
}

// skip to next directory entry (i.e. file)
unsigned int get_next_file_addr(DIR_t* dir)
{
	return dir->this_addr+0x20;	// skip past this entry
}

// is dir a directory?
static unsigned char is_dir(DIR_t* dir)
{
	return (dir->attributes == 0x10);
}

// read in necessary information to dir
void load_dir(DIR_t *dir, unsigned int cluster_address)
{
	unsigned char		buf[32];		// the size of a dir
	unsigned short		cluster_no_low2;
	unsigned short		cluster_no_high2;


	// the entries that are used for the long file name info have attributes == 0x0f
	// let's skip those, shall we?
	// The same goes with the volume label. We don't care about the volume label,
	// but if it exists we must deal with it. Solution: skip it :)
	// Also, if the file has been deleted, we will search for the next one
	// we will stop searching if... ehh... hmm
	// actually we know that the last entry is found when attribute == 0x00
	// but we don't use that. Haha!
	
	cluster_address = cluster_address - 0x20;

	do {
		cluster_address = cluster_address + 0x20;

		// go to file
		seek_from_start(cluster_address);
		// read 32 bytes into buffer
		read_from_io(buf, 1, 32);

		dir->attributes = buf[0x0b];
		my_memcpy(dir->filename, buf, 8);

	} while (dir->attributes == 0x0f || dir->attributes == 0x08 || dir->filename[0] == 0xe5 ||
			dir->filename[0] == 0x2e);

	dir->this_addr = cluster_address;
	my_memcpy(dir->extension, buf+0x08, 3);
	cluster_no_high2 = load_short(buf+0x14);
	cluster_no_low2 = load_short(buf+0x1a);
	dir->first_cluster = ((unsigned int)cluster_no_high2)<<16 | (unsigned int)cluster_no_low2;
	dir->file_size = load_int(buf+0x1c);
}

// print the directory entry info to screen
void print_dir(DIR_t *dir)
{
	xil_printf("--- Print dir ---\r\n");
	int		i;

	xil_printf("Address: 0x%08x\r\n", dir->this_addr);
	xil_printf("First byte: 0x%02x\r\n", dir->filename[0]);
	xil_printf("Filename: ");
	for (i = 0; i < 8; i++) {
		xil_printf("%c", dir->filename[i]);
	}
	xil_printf(".");
	for (i = 0; i < 3; i++) {
		xil_printf("%c", dir->extension[i]);
	}
	xil_printf("\r\n");
	xil_printf("Attributes: 0x%02x\r\n", dir->attributes);
	xil_printf("First cluster no: 0x%08x\r\n", dir->first_cluster);
	xil_printf("File size: %u\r\n", dir->file_size);
}

// copy a whole file to dest. size will have to be checked before so that the file fit in dest
void copy_file(void* dest, unsigned int cluster_no, unsigned int size)
{
	xil_printf("\r\n--- Copy file ---\r\n");
	xil_printf("Cluster no: 0x%08x\r\n", cluster_no);
	
	unsigned char		keep_on;
	unsigned int		next_cluster;
	unsigned int		remaining_size;
	unsigned int		sector_address;
	unsigned int		sector_count;
	unsigned char		sector_buf[512];
	unsigned int		i;
	unsigned int		j;
	
	remaining_size = size;
	sector_count = 0;
	keep_on = 1;
	
	next_cluster = cluster_no;
	//at what address does the files first cluster (and as well first sector) start?
	sector_address = get_cluster_addr(next_cluster);
	
	while (keep_on) {
		// only print every 10th sector that will be read, otherwise it will slow down
		if (sector_count % 10 == 0) {
			xil_printf("Sector count: %d\r\n", sector_count);
		}
		
		// go to the right block
		seek_from_start(sector_address);
		
		// if remaining size > bytes per sector, the we have to read all 512 bytes,
		// if not, then we only have to read remaining size
		if (remaining_size > fat32.bytes_per_sector) {
			read_from_io(dest+fat32.bytes_per_sector*sector_count, 1, 512);
			remaining_size -= 512;
		} else {
			read_from_io(dest+fat32.bytes_per_sector*sector_count, 1, remaining_size); 
			remaining_size = 0;
			keep_on = 0;
		}
		
		sector_count++;
		sector_address += 512;
		
		// do we have to switch cluster to read sectors from?
		if (sector_count % fat32.sectors_per_cluster == 0) {
			next_cluster = get_next_cluster(next_cluster);
			if (next_cluster >= 0x00000002 && next_cluster <= 0x0fffffef) {
				sector_address = get_cluster_addr(next_cluster);
			} else {
				keep_on = 0;
			}			
		}
	}
}

// print all files in dir and its sub directories
void print_file_list(DIR_t *dir)
{
	xil_printf("\r\n--- List files ---\r\n");

	unsigned int		dir_address;
	DIR_t			tmp;
	DIR_t			dir_tmp;

	tmp = *dir;

	while (tmp.first_cluster > 0x00000002 && tmp.first_cluster <= 0x0fffffef) {
		print_dir(&tmp);

		if (tmp.attributes == 0x10) {
			dir_address = get_cluster_addr(tmp.first_cluster);
			load_dir(&dir_tmp, dir_address);
			print_file_list(&dir_tmp);
		}

		// find the next existing file
		// 0xef means it has been deleted
		do {
			dir_address = get_next_file_addr(&tmp);
			load_dir(&tmp, dir_address);
		} while (tmp.filename[0] == 0xe5);
	}
}

void create_file_list(DIR_t* dir, list_t* list)
{
	unsigned int		dir_address;
	DIR_t					tmp;
	DIR_t					dir_tmp;

	tmp = *dir;

	xil_printf("Creating file list\r\n");
	while (tmp.first_cluster >= 0x00000002 && tmp.first_cluster <= 0x0fffffef) {
		//xil_printf("tmp.attributes = %x \r\n",tmp.attributes);
		//xil_printf("tmp.attributes = %d \r\n",tmp.attributes);
		if (tmp.attributes == 0x10) {  // 10
		
			dir_address = get_cluster_addr(tmp.first_cluster);
			load_dir(&dir_tmp, dir_address);
			create_file_list(&dir_tmp, list);
			xil_printf("Creating file list inside if  list->next = %d\r\n ",list->next);
			
		} else {
			list_add(list, tmp.first_cluster, tmp.file_size, tmp.filename, tmp.extension);
		}

		// find the next existing file
		// 0xe5 means it has been deleted
		do {
			dir_address = get_next_file_addr(&tmp);
			load_dir(&tmp, dir_address);
		} while (tmp.filename[0] == 0xe5);
	}
}
