/*
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _fat32_h_
#define _fat32_h_

#include "list.h"

struct FAT32_t {
	/* General Volume Boot Record */
	unsigned short		bytes_per_sector;
	unsigned char		sectors_per_cluster;
	unsigned short		reserved_sector_count;
	unsigned char		number_of_fats;		// how many tables? Always 2
	unsigned short		max_root_dir_entries;	// max number of entries in root dir. Shoudl be 0 for fat32
	
	unsigned short		sectors_per_fat;
	unsigned int		fat_size;		// sectors_per_fat*number_of_fats
	unsigned int		root_dir_sector;	// what sector does root dir start at?
	unsigned char		fat_type[8];
	unsigned short		signature;

	unsigned int		fat_base;		// fat start sector
	unsigned int		data_start_addr;	// data start address
} fat32;


typedef struct DIR_t DIR_t;

/* Represents a directory entry, a file */
struct DIR_t {
	unsigned char		filename[8];
	unsigned char		extension[3];
	unsigned char		attributes;
	unsigned int		this_addr;		// the address cluster for the DIR_t
	unsigned int		first_cluster;		// cluster number for the first part of the first file
	unsigned int		file_size;		// size of the first file
};


void init_vbr();
void print_vmbr();

// transform cluster number to cluster address
unsigned int get_cluster_addr(unsigned int cluster_no);
// search for next file address
unsigned int get_next_file_addr(DIR_t* dir);

// get next cluster number from fat table
unsigned int get_next_cluster(unsigned int cluster_no);

// read from file and put in dir
void load_dir(DIR_t *dir, unsigned int cluster_address);
// print out all members
void print_dir(DIR_t *dir);

// copy file at cluster_o to dest
void copy_file(void* dest, unsigned int cluster_no, unsigned int size);

// print all files in dir and its sub directories
void print_file_list(DIR_t *dir);

// create a list of the cluster numbers for the file
void create_file_list(DIR_t* dir, list_t* list);

#endif
