/*
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
   */


#include "stdlib.h"
#include "stdio.h"
#include "xbasic_types.h"
#include "xio.h"
#include "SD_Driver.h"
#include "io.h"

// we need this because we read block by block
static unsigned char		blockbuf[512];
static unsigned int		last_used_block;
static unsigned int		block;
static unsigned int		offset;


extern  Xboolean gbActive;

void init_io()
{
	last_used_block = 0xffffffff; // this will never happen
	xil_printf("Init io...\r\n");
	
	unsigned char first_time = 1;
	Xuint32 i,j;

	while(!SD_Init()) {	  
		if (first_time) {
			xil_printf("Please insert SD card\r\n");
			first_time = 0;
		} 
				  
		for (j=0;j<10000;j++)
			for (i=0;i<100;i++)
				asm("nop");  // delay				
	}
	
	
	xil_printf("SD found!\r\n");
}

void seek_from_start(unsigned int address)
{
	block = address/512;
	offset = address%512;
}

void read_from_io(void* dest, unsigned int size, unsigned int count)
{
	unsigned int		count_remaining;
	unsigned char		keep_on;
	unsigned int		current_block;
	unsigned int		current_offset;
	unsigned int		remaining_count;
	unsigned int		counter;
	Xboolean				result;

	
	counter = 0;
	current_block = block;
	current_offset = offset;
	remaining_count = count;

	//printf("Reading\t\t size: 0x%08x  count: 0x%08x\n", size, count);


	keep_on = 1;
	while (keep_on) {		
		// it's unneccessary to read the block if it's already in the block buffer
		if (current_block != last_used_block) {
			result = SD_read_block(current_block, blockbuf);
			last_used_block = block; // store away, because if it's the same we don't have to do a re-read
			
			if (result == XFALSE){
				xil_printf("Read failure from block: %08x\r\n", current_block);
			}
		}
	
		
		// do we have to read another sector?
		if (current_offset+remaining_count > 512) {
			my_memcpy(dest+counter*512, blockbuf+current_offset, 512-current_offset);
			current_offset = 0;
			remaining_count -= (512-current_offset);
			current_block++;
			//printf("----------- RESULT ---------------\n");
			//printf("%s", dest);
		} else {
			// we don't have to read another sector
			my_memcpy(dest+counter*512, blockbuf+current_offset, remaining_count);
			keep_on = 0;
		}
		counter++;
	}	
}

void close_io()
{
	
}

void my_memcpy(void *dest, void *src, unsigned int count)
{
	char            *dest_c;
	char            *src_c;
	int             i;

	dest_c = (char*)dest;
	src_c = (char*)src;

	for (i = 0; i < count; i++) {
		//dest_c[i] = src_c[i];
		XIo_Out8(dest_c+i, XIo_In8(src_c+i));
	}
}

